/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaInputQa.h
/// \date   01.03.2024
/// \brief  A QA module for CA tracking input data (header)
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "CaEnumArray.h"
#include "CaQaBase.h"

namespace cbm::algo::ca
{
  /// \class cbm::algo::ca::qa::InputQa
  /// \brief InputQa class for the CA tracking QA (header)
  ///
  class InputQa : public QaBase {
    /// \brief Hit set entries
    enum class EHitSet
    {
      Input,  ///< Input hits
      Used,   ///< Hits used in tracks
      kEND
    };

    /// \brief Definition of enum array over EHitSet entries
    template<typename T>
    using HitSetArray_t = EnumArray<EHitSet, T>;

    /// \brief Array of EHitSet entries for iteration
    static constexpr HitSetArray_t<EHitSet> kHitSets = {EHitSet::Input, EHitSet::Used};

   public:
    /// \brief Default destructor
    /// \param pSender  Pointer to the histogram sender
    InputQa(std::shared_ptr<HistogramSender> pSender) : QaBase(pSender, "CA/Input"){};

    /// \brief Constructor from the configuration object
    /// \param config  QA configuration object
    InputQa() = default;

    /// \brief Copy constructor
    InputQa(const InputQa&) = delete;

    /// \brief Move constructor
    InputQa(InputQa&&) = delete;

    /// \brief Destructor
    ~InputQa() = default;

    /// \brief Copy assignment operator
    InputQa& operator=(const InputQa&) = delete;

    /// \brief Move assignment operator
    InputQa& operator=(InputQa&&) = delete;

    /// \brief QA execution function
    void Exec();

    /// \brief Initializes the QA
    void Init();

   private:
    /// \brief Fills hit distributions
    /// \param  hitSet  Hit set enum entry
    /// \param  hit     Reference to hit
    void FillHitDistributionsForHitSet(EHitSet hitSet, const ca::Hit& hit);

    static constexpr double kXYZMargin = 0.05;  ///< Margin for occupancy distributions in XY plane
    static constexpr int knHitSets     = 2;     ///< Number of hit sets: input/used

    // Hit distributions vs. station
    using OccupHistContainer_t = std::vector<HitSetArray_t<qa::H2D*>>;
    OccupHistContainer_t fvphHitOccupXY;  ///< hist: Hit occupancy in different stations in XY plane
    OccupHistContainer_t fvphHitOccupZX;  ///< hist: Hit occupancy in different stations in ZX plane
    OccupHistContainer_t fvphHitOccupZY;  ///< hist: Hit occupancy in different stations in ZY plane

    std::vector<qa::Prof2D*> fvphHitUsageXY;  ///< prof: Hit usage in different stations in XY plane

    HitSetArray_t<qa::H1D*> fvphHitFrontKeyIndex = {nullptr, nullptr};  ///< Indices of front hit keys
    HitSetArray_t<qa::H1D*> fvphHitBackKeyIndex  = {nullptr, nullptr};  ///< Indices of back hit keys
  };
}  // namespace cbm::algo::ca
